/************************************************************************/
/*                                                                      */
/*   Program Name :  k5tom5b                                            */
/*                                                                      */
/*     K5tom5b is a K5 to Mark5B file format converter                  */
/*                                                                      */
/*     Enviroment variables                                             */
/*           M5DIR      ---   default directory for Mark5 data          */
/*           M5VEX      ---   default directory for Mark5 vex file      */
/*                                                                      */
/*            Ver 1.70  2016-08-15                                      */
/*                                                                      */
/*      Copyright (c) 2009-2016   T.Kondo/NICT All Right Reserved       */
/*                                                                      */
/************************************************************************/
/*
     2009-06-07  -- Ver 1.00  start development based on "k5tom5"
     2009-07-31  -- Ver 1.10  debug in case of that start time differs amoung K5 files.
                              This bug was taken over from "k5tom5"
     2010-07-21  -- Ver 1.20  update so as to guarantee that integer second of data length
     2010-08-03  -- Ver 1.30  add option "-nocheck" to ignore header error
                              for 2nd and later headers (1st header is checked to
                              obtain sampling information)
     2011-01-09  -- Ver 1.31  update time counting more accurate to avoid 
                              time error in case of 64Mspsx2bit mode
     2012-12-03  -- Ver 1.32  debug in the calculation of common_sec_length
                    Ver 1.40  debug in the bug in the case of 64Mhz x 2bit x 4ch case
                              (this is considerably big debug)
     2014-05-24  -- Ver 1.50  add "-chN" and "-chall" option to swtch channel assign freely
     2014-06-11  -- Ver 1.51  precise command length check was added
     2016-07-20  -- Ver 1.60  debug when day change occurs during a scan
     2016-08-15  -- Ver 1.70  revised to allow that 1st run parameter can be omiited or 0 for K5/VSSP32 data


   how to run
          k5tom5b jjj|yyyymmdd|yyyyddd 
                        k5name1 [k5name2 k5name3 k5name4 [m5name]] [options]

            or

          k5tom5b  k5name1 [k5name2 k5name3 k5name4 [m5name]] [options]   <= for K5/VSSP32
          

    
    where    
             jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year
                       or
                       0      : for K5/VSSP32 data (because date info is included in VSSP32 format) <=2016-08-15

                        of observation date
             
             k5name1 -- K5 data file name containing Ch01-04 or Ch01
             k5name2 -- K5 data file name containing Ch05-08 or Ch02
             k5name3 -- K5 data file name containing Ch09-12 or Ch03
             k5name4 -- K5 data file name containing Ch13-16 or Ch04
                  2nd and after K5 file names can not be accompanied by 
                  directory name, the directory of 1st K5 file is 
                  adopted for these files

             m5name  -- Mark5 file name to be created

             options (any order)
               -o m5name     -- Mark5 file name to be created
                                 (default : see below)
               -d m5dir      -- directory for Mark5 file out
                                 (default  : mark5 directory)
               -s strat_sec  -- offset time from data head (sec)
                                 (default :  0 )
               -p period     -- data period to convert (sec)
                                 (default : all)

               -date jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year

               -nocheck      -- no header error check (adopted for 2nd and later headers)

                **** follwoing will be implemented *****

               -ch1 nn       -- set K5 channel corresponding to Mark5 ch #1
                                     nn is K5 channel number (1-16)
               -ch2 nn       -- set K5 channel corresponding to Mark5 ch #2
                   ...........................
               -ch16 nn      -- set K5 channel corresponding to Mark5 ch #16

               -chall n1:n2:....:n16    -- set K5 channels corresponding to Mark5 all channels
                               n1  -- K5 channel number (1-16) corresponding to Mark5 ch 1
                               n2  -- K5 channel number (1-16) corresponding to Mark5 ch 2
                                 ...............
                               n16 -- K5 channel number (1-16) corresponding to Mark5 ch 16



  Enviroment variables
          M5DIR    ---   default path to Mark5 data
          M5VEX    ---   default path to Mark5 vex file


  AD bit is limited to 2bit max. 


*** sample run parameters ****
*/

#define PROGRAM_VERSION "k5tom5b (Ver 1.70  2016-08-15)"

//#define ORIGINAL    // for debug

#include "libk5.h"      // common header for IP-VLBI utilities
#include "libmk5tok5.h"  // library for Mark5 to K5 conversion
#include "libvssp32.h"  // library for Mark5 to K5 conversion

/* default conversion information file name set */
#define INFONAME_DFLT "k5tom5binfo.txt"   // information file default name
/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory
/* environment variable for Mark5 vex file default directory set */
#define M5VEX "M5VEX" // the environment variable for Mark5 vex file

/* default parameters  */
#define K5NUMCH  4
#define M5PARITY 0
#define M5VLBA   0
#define M5FANOUT 1
#define M5ADBIT  1
#define M5SAMPLE 4   // default sampling frequency in MHz
#define M5NTRACK 32

//#define HUYGENS  // Huygens processing mode

int cmnd_get_n_p(int argc, char *argv[], char *cmnd, char *param)
// do not care about upper case and lower case 
// but # of characters are precisely compared

// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];
    char command[80];

    param[0]='\0';

    // convert to lower case character
    strcpy(command,cmnd);
    conv2small(command);
    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        conv2small(cbuf); // convert to small characters
        if(strncmp(cbuf,command,strlen(command))==0 && strlen(cbuf)==strlen(command)){  
            n++;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
            return 0;
        }
        n++;
    }
    return -1;
}


int check_K5file_or_not(char *fname)
{
/* 
      check K5 file or not
      2016-08-15

      return
           -99: file not existed
           -1: not K5 file 
            1: K5/VSSP
            2: K5/VSSP32
            3: K5/VSSP64

*/
    int flag_open;
    int i,iret;
    int maxbytes=256;  // read max bytes to judge file contents
    K5VSSP_file_t *k5;   // K5 data type variables 

    flag_open=0;
    iret=-1;
    k5=(K5VSSP_file_t *)malloc(sizeof(K5VSSP_file_t));
    k5vssp_t_ini(k5);
    //strcpy(k5->fname,fname);
    if( (k5->stream  = fopen( fname, "rb" )) == NULL ){
        //printf("Data File %s   Open Error! \n",fname);
        iret=-99;
        goto exit2return;
    }
    flag_open=1;
    //k5->size=get_file_size(fname);
    /************************************************/
    /* Data header check to judge file contents     */
    /************************************************/
    if(CheckK5header_with_Read2(k5,maxbytes) < 0){
        goto exit2return;
    }
    iret=k5->format;
exit2return:
    if(flag_open==1) i=fclose(k5->stream);
    free(k5);
    return iret;
}



void param_get_k5tom5b(int argc, char *argv[], k5m5runp_t *runp)                       
{
    int n,i;
    int len;
    int mode;
    char cbuf[256],buf[10];
    //double dw; // work
    vparam_t vp;

    // initialize parameters

    /* common parameters */
    runp->ptype=1;  // run with information file
    for(n=0; n<4; n++) runp->k5name[n][0]='\0';
    runp->mk5name[0]='\0';

    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->reverse=0;
    runp->makemode=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
    runp->dflt_pass[0]='\0';

    /* mode 2 parameters */
    runp->numch=K5NUMCH;
    runp->ntrack=M5NTRACK;
    runp->adbit =M5ADBIT;
    runp->parity=M5PARITY;
    runp->nrzm  =-1;
    runp->fanout=M5FANOUT;
    runp->vlba  =M5VLBA;
    runp->tot_track=0;
    runp->fmhz  =(double)M5SAMPLE;
    runp->e2bit=2;
    for(n=0; n<32; n++) runp->ibit[n]=0;

    runp->yyyy=0; runp->mm=0; runp->dd=0; runp->ddd=0;
    runp->mjd=0; runp->mjd3=0;

    runp->nbits=32;
    runp->modulation=0;  // modulation off 2008-07-04

    runp->nocheck=0;  // 2010-08-03
	for(n=0; n<16; n++) runp->ch_assign[n]=n+1;   // 2014-05-22

    mode=0;
    /* check 1st param whether it is started with "-" or not  */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        if(cbuf[0] == '-') mode=1;
    }

    /* check 1st parama whether it is K5 file or not */ // 2016-08-15
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        if(i=check_K5file_or_not(cbuf) > 0){   // K5/VSSP file
            n=2;
            goto k5_name_get;
        }
    }


    if(mode == 1) {
        n=2;
        goto comm_anal;
    }

    /* 1st param should be either jjj(3digit MJD) or yyyymmdd or yyyyddd */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        len=strlen(cbuf);
        //printf("len = %d\n",len);
        if(len==8){ // yyyymmdd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,2); buf[2]='\0';
            runp->mm=atoi(buf);
            strncpy(buf,cbuf+6,2); buf[2]='\0';
            runp->dd=atoi(buf);
            //get MJD
            runp->mjd=ut2mjd(&runp->dw,runp->yyyy,runp->mm,runp->dd,0,0,0);
            //get total day 
            md2ddd(runp->yyyy,&runp->ddd,runp->mm,runp->dd);
        } else if(len==7){ // yyyyddd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,3); buf[3]='\0';
            runp->ddd=atoi(buf);
            //get month day
            ddd2md(runp->yyyy,runp->ddd,&runp->mm,&runp->dd);
            //get MJD
            runp->mjd=ut2mjd(&runp->dw,runp->yyyy,runp->mm,runp->dd,0,0,0);
        } else {
            runp->mjd = atoi(cbuf);
        }
        runp->mjd3 = runp->mjd % 1000;
    }

    n=3;

k5_name_get:     // 2016-08-15
    runp->k5num=0;

    /* K5 name get */
    while(1){
        if(argc >= n) {
           strcpy(cbuf,*(argv+ n-1) );
           //printf("%s\n",cbuf);
           if(cbuf[0] == '-') break;  // go to command parameter get
           //if(runp->k5num >= 4) {
           //    printf(" Warning!! Number of K5 file exceeds 4!!\n");
           //    printf(" So more K5 files will be neglected!!\n");
           //    runp->k5num=4;
           //    n++;
           //    break;
           //}
           if(runp->k5num >= 4) {  // mark5 name
               strcpy(runp->mk5name,cbuf);
               runp->k5num=4;
               n++;
               break;
           }
           strcpy(runp->k5name[runp->k5num],cbuf);
           runp->k5num++;
        } else {
            goto exitend;
        }
        n++;
    }


comm_anal:

	if(cmnd_get_n(argc,argv,"-o",cbuf)==0){     // Mark5B name
		strcpy(runp->mk5name,cbuf);
        /* check Mark5 file name is acompanied by directory or not */
        get_true_path(runp->mk5name, M5DIR, "./",cbuf);
        strcpy(runp->mk5name,cbuf);
	}
	if(cmnd_get_n(argc,argv,"-d",cbuf)==0){     // out directory
        strcpy(runp->outdir,cbuf);
        /* env variable is substituted by real one */
        if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
            strcpy(runp->outdir,cbuf);
        }
    }
    if(cmnd_get_n(argc,argv,"-e",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2 bit encode mode
    if(cmnd_get_n(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf); // start offset
    if(cmnd_get_n(argc,argv,"-p",cbuf)==0) runp->period = atoi(cbuf);  // period
    if(cmnd_get_n(argc,argv,"-nocheck",cbuf)==0) runp->nocheck=1;      // no check mode
    if(cmnd_get_n(argc,argv,"-mon",cbuf)==0) runp->monit=1;            // monitor mode on
    if(cmnd_get_n_p(argc,argv,"-ch1",cbuf)==0) runp->ch_assign[0]=atoi(cbuf);  // CH1 set  2014-06-11
    if(cmnd_get_n(argc,argv,"-ch01",cbuf)==0) runp->ch_assign[0]=atoi(cbuf);  // CH1 set
    if(cmnd_get_n(argc,argv,"-ch2",cbuf)==0) runp->ch_assign[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch02",cbuf)==0) runp->ch_assign[1]=atoi(cbuf);  // CH2 set
    if(cmnd_get_n(argc,argv,"-ch3",cbuf)==0) runp->ch_assign[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch03",cbuf)==0) runp->ch_assign[2]=atoi(cbuf);  // CH3 set
    if(cmnd_get_n(argc,argv,"-ch4",cbuf)==0) runp->ch_assign[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch04",cbuf)==0) runp->ch_assign[3]=atoi(cbuf);  // CH4 set
    if(cmnd_get_n(argc,argv,"-ch5",cbuf)==0) runp->ch_assign[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch05",cbuf)==0) runp->ch_assign[4]=atoi(cbuf);  // CH5 set
    if(cmnd_get_n(argc,argv,"-ch6",cbuf)==0) runp->ch_assign[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch06",cbuf)==0) runp->ch_assign[5]=atoi(cbuf);  // CH6 set
    if(cmnd_get_n(argc,argv,"-ch7",cbuf)==0) runp->ch_assign[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch07",cbuf)==0) runp->ch_assign[6]=atoi(cbuf);  // CH7 set
    if(cmnd_get_n(argc,argv,"-ch8",cbuf)==0) runp->ch_assign[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch08",cbuf)==0) runp->ch_assign[7]=atoi(cbuf);  // CH8 set
    if(cmnd_get_n(argc,argv,"-ch9",cbuf)==0) runp->ch_assign[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch09",cbuf)==0) runp->ch_assign[8]=atoi(cbuf);  // CH9 set
    if(cmnd_get_n(argc,argv,"-ch10",cbuf)==0) runp->ch_assign[9]=atoi(cbuf);  // CH10 set
    if(cmnd_get_n(argc,argv,"-ch11",cbuf)==0) runp->ch_assign[10]=atoi(cbuf);  // CH11 set
    if(cmnd_get_n(argc,argv,"-ch12",cbuf)==0) runp->ch_assign[11]=atoi(cbuf);  // CH12 set
    if(cmnd_get_n(argc,argv,"-ch13",cbuf)==0) runp->ch_assign[12]=atoi(cbuf);  // CH13 set
    if(cmnd_get_n(argc,argv,"-ch14",cbuf)==0) runp->ch_assign[13]=atoi(cbuf);  // CH14 set
    if(cmnd_get_n(argc,argv,"-ch15",cbuf)==0) runp->ch_assign[14]=atoi(cbuf);  // CH15 set
    if(cmnd_get_n(argc,argv,"-ch16",cbuf)==0) runp->ch_assign[15]=atoi(cbuf);  // CH16 set

    if(cmnd_get_n(argc,argv,"-chall",cbuf)==0) { // all channel set
        vex_parse(cbuf, &vp);
        i=0;
        for(n=0; n<vp.n; n++) {
           //printf(" n= %d  %s\n",n,vp.param[n]);// debug
            if(vp.param[n][0]!=':'){
                runp->ch_assign[i]=atoi(vp.param[n]);
                i++;
            }
        }

    }

exitend:;

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("How to use\n");
    printf("\n");
    //printf("    k5tom5 yymmdd k51name [k52name k53name k54name] [options]\n");
    printf("    k5tom5b jjj|yyyymmdd|yyyyddd\n");
    printf("                   k5name1 [k5name2 k5name3 k5name4 [m5name]] [options]\n");
    printf("        or\n");
    printf("    k5tom5b k5name1 [k5name2 k5name3 k5name4 [m5name]] [options] (for K5/VSSP32)\n");
    //printf("    k5tom5b -i -infofile|make -v vexfile [options]\n");
    //printf("               .... for information file creation mode\n");
    //printf("\n");
    //printf("where  yymmdd ---- year(2digits),month(2digits),day of month (2digits) of\n");
    //printf("                      data start time\n");
    printf("   where\n");
    printf("       jjj|yyyymmdd|yyyyddd  -- observation date in three ways as follows\n");
    printf("           jjj      : lower 3digits of MJD \n");
    printf("              or\n");
    printf("           yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month\n");
    printf("              or\n");
    printf("           yyyyddd  : 4 digit year, 3 digit day of year\n");
    printf("              or\n");
    printf("           0 or omitted : in case of the use of K5/VSSP32 file(s)\n");
    printf("       k5name1 ---- K5 file name for CH01-04 or CH01\n");
    printf("       k5name2 ---- K5 file name for CH05-08 or CH02\n");
    printf("       k5name3 ---- K5 file name for CH09-12 or CH03\n");
    printf("       k5name4 ---- K5 file name for CH13-16 or CH04\n");
    printf("                      2nd and after K5 file names are not accompanied by\n");
    printf("                      directory name, the directory of 1st K5 file is \n");
    printf("                      adopted for these files\n\n");
    printf("       m5name  ---- Mark5 file name to be created\n\n");
    printf("      options (any order)\n");
    //printf("        -i info_file  -- infomation file name that contains\n");
    //printf("                           track vs channel table\n");
    //printf("                           bit_psition vs channel table\n");
    //printf("                           group# vs channel table\n");
    //printf("                           (default \"%s\")\n",INFONAME_DFLT);
    //printf("                           \"-i -info_file\" means\n");
    //printf("                             create infomation file named info_file\n");
    //printf("                           \"-i make\" means create default info file\n");

    printf("        -o m5name   -- Mark5 file name to be created\n");
    //printf("                          (default : see below)\n");
    printf("        -d m5dir    -- directory for Mark5 file out\n");
    printf("                          (default  : mark5 directory)\n");
    printf("        -s soffset  -- set offset time from data head (sec)\n");
    printf("                          (default :  0 )\n");
    printf("        -p period   -- data period to convert (sec)\n");
    printf("                          (default : all)\n");
    printf("        -e2bit n    -- 2bit data encode mode\n");
    printf("                        n=1 : offset binary 0 1 2 3\n");
    printf("                        n=2 : twisted offset binary 0 2 1 3\n");
    printf("                        n=3 : signed binary 1 0 2 3\n");
    printf("                          (default is 2)\n");
    printf("        -nocheck    -- no header error check\n"); 
    printf("                          (adopted for 2nd and later headers)\n"); 
    printf("        -ch1 nn     -- set K5 channel corresponding to Mark5 ch #1\n"); 
    printf("                          nn is K5 channel number (1-16)\n"); 
    printf("        -ch2 nn     -- set K5 channel corresponding to Mark5 ch #2\n"); 
    printf("           ...........................\n"); 
    printf("        -ch16 nn    -- set K5 channel corresponding to Mark5 ch #16\n"); 
    printf("        -chall n1:n2:....:n16\n");
    printf("               -- set K5 channels corresponding to Mark5 all channels\n"); 
    printf("                   n1  -- K5 channel number (1-16) corresponding to Mark5 ch 1\n"); 
    printf("                   n2  -- K5 channel number (1-16) corresponding to Mark5 ch 2\n"); 
    printf("                    ...............\n"); 
    printf("                   n16 -- K5 channel number (1-16) corresponding to Mark5 ch 16\n"); 
    //printf("        -mark4        -- set frame format to Mark4 mode (default VLBA mode)\n");
    //printf("        -vlba         -- set frame format to VLBA mode (default Mark4 mode)\n");
    //printf("        -long         -- set 64 bit(track) mode (default 32bit mode)\n");
    //printf("        -v vex_file   -- VEX file name to be read when \"information\n");
    //printf("                              file creation mode\" is set\n");
    //printf("        -sid stat_id  -- station ID when \"information file creation\n");
    //printf("                              mode\" is set\n");
    //printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
    //printf("                              creation mode\" is set (default : 1)\n");
    //printf("        -monit        -- info_file monitor ON\n");
    printf("\n");
}

void monit_runp(k5m5runp_t *runp)
// run parameter monitor
{
    int n;

       printf(" ---------- RUN CONDITION ----------\n");
       printf("   yyyymmdd  : %04d%02d%02d\n",runp->yyyy,runp->mm,runp->dd);
       printf("   yyyyddd   : %04d%03d\n",runp->yyyy,runp->ddd);
       printf("   jjj       : %03d\n",runp->mjd3);
       if(runp->yyyy != 0){
       printf("   MJD       : %d\n",runp->mjd);
       }
       for(n=0; n<runp->k5num; n++)
       printf("   k5name %d  : %s\n",n+1,runp->k5name[n]);

       printf("   mk5name   : %s\n",runp->mk5name);
       printf("   outdir    : %s\n",runp->outdir);
       //printf("   infofile  : %s\n",runp->info_file);
       //printf("   channel   : %d\n",runp->channel);
       //printf("   group     : %d\n",runp->group);
       printf("   soffset   : %d\n",runp->soffset);
       printf("   period    : %d\n",runp->period);
       printf("   e2bit     : %d\n",runp->e2bit);
       printf("   nocheck   : %d\n",runp->nocheck);
       printf("     ----- channel assign -----\n");
       printf("     M5b CH#  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16\n");
       printf("     K5  CH#");
       for(n=0; n<16; n++) printf("%3d",runp->ch_assign[n]);
       printf("\n");
       //printf("   odd       : %d\n",runp->odd);
       //printf("   parity    : %d\n",runp->parity);
       //printf("   nrzm      : %d\n",runp->nrzm);
       //printf("   vlba      : %d\n",runp->vlba);
       //printf("   nbits     : %d\n",runp->nbits);
       //if(runp->vex_file[0] != '\0')
       //   printf("   vex_file  : %s\n",runp->vex_file);
       //if(runp->sid[0] != '\0' && runp->sid[0] !='-')
       //   printf("   station_id: %s\n",runp->sid);
       //printf("   modulation: %d\n",runp->modulation);
       printf(" -------------------------------------\n");

    printf("\n");
}

void setup_k5files_full_path(k5m5runp_t *runp)
{    // 2008-09-30
/*
   full path of K5 filename is setupped using the directory
   of 1 st K5 name if 2nd and after K5 files are not acompanied 
   by directory

*/
    char dir0[128],dir[128],fname[128];
    int n;
    parse_dir_and_name(runp->k5name[0], dir0, fname);
    if(dir0[0]!='\0'){
        for(n=1; n<runp->k5num; n++){
            parse_dir_and_name(runp->k5name[n], dir, fname);
            if(dir[0]=='\0'){
               strcpy(runp->k5name[n],dir0);
               strcat(runp->k5name[n],fname);
            }
        }
    }
}

void parseADdataKOKO(u_int32_t *ia, int numb, int adbit, int *xdata, int *npos)
{
/*
    IP-VLBI{[hf[^A/D\ɉTvOlɕϊ
      ch͂̎_ł͊֌WȂ

      [in]
         ia  ---  sampler (IP-VLBI board) out data 
         numb --  # of elements of ia 
         adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         xdata -- data parsed into each sampled data
         npos  -- valid number of data in xdata array

     Relation between # of ch (numch) and each ch time siries

        chnumchƂ
        ch# mk Ԗڂ̃Tvf[^

        sdata(m,k)=xdata(k*numch+m)

        ƂȂ
        
*/
     int numb32, pat, n, i, ilast, ipos, ww;
     u_int32_t wd1,wd2;
     //numb32=numb/4;
     numb32=numb;
     ipos=0;
     ilast=32/adbit;
     pat=255;
     if (adbit == 1) {pat=1;}
     if (adbit == 2) {pat=3;}
     if (adbit == 4) {pat=15;}
     // if (adbit == 8) {pat=255;} 

     //for( n=0; n < numb32-1; n++)
     for( n=0; n < numb32; n++)
     {
        wd1=ia[n];
        wd2=ia[n+1];
        /* check sync pattern */
        //if ( (wd1 == 0xFFFFFFFFUL) && ((wd2 & 0xFF000000UL) == 0xFF000000UL) )  // sync detected
//        if ( (wd1 == 0xFFFFFFFFUL) && ((wd2 & 0xFF000000UL) == 0x8B000000UL) )  // sync detected 02-07-17
//         {      //printf("Sync detected ");   // removed  on 2002-10-3 
//                goto exit1;
//         }
        for(i=0; i<ilast; i++)
        {
             ww=(int)(wd1 & pat);
             xdata[ipos]=ww;
             wd1=wd1 >> adbit ; 
             ipos++;
         }
      // exit1:  ;
     }
     *npos=ipos;
}


int main(int argc, char *argv[])
{
       int i,k,n,m;
       int ch;
       int aflag;
       int secmin,secmax;
       int iread,iwrite;
       int kpos;
       unsigned int iw;
       //int obufmax=225;
       int ow;

       /* for Mark 5B format */ 
       unsigned int obuf[2504];  // for 1 Frame data 2504 X 32 bits
       unsigned int *head;   // for header 4 X 32 bits
       unsigned int *idata;  // for data field 2500 X 32 bits
       int uid;   // lower 16 bits are valid
       
       int nbuf=2504;   // @
       int frame32=2500;   // # of 32-bit words in a Mark5B frame
       int tvgflag;
       int mjd, seconds, nframe;
       int msec10;    // fractional second in 0.1 msec unit
       int seconds_init; // start time in seconds 2011-01-09

       int totch;  // total number of channels
       int bitsample;  // # of bits/sample
       int nsample32;   // # of samples per 32 bits

       /* ---------- */



       int period;

       int sfreq, adbit, numch;
       int imax, isa;
       int usampl;
       int samplpp;  // samples/PP
       int bits, bytes, numb;
       int tmax,tmin;
       int FileCloseFlag[4];  // 2008-07-04  no more use 2016-08-15
       int K5_FileOpenFlag[4];  // 2016-08-15
       int M5_FileOpenFlag;     // 2016-08-15
       //int bittable[64];
       double fsampl,tsampl,unitp;
       double sec_period[4];  // contain actual data size for 4 files (in sec)
       int isec_period[4];
       int common_sec_length;
       int adjust_flag;   // flag to set conversion length integer second  1:set 0: no set
       double dwork;
       double dmsec10;
       double dsec;  // 2011-01-09
       int sec_count; // 2012-12-03
       
       int K5file_id[16];  // to show file id (by 1-4) correspond to K5 ch# 1-16   (2014-05-23)
       int K5chan_id[16];  // to show chanel # in K5 file (by 1-4) correspond to K5 ch# 1-16   (2014-05-23)
       int f_id,ch_id;

       double telapsed;           // to measure processing time
       char fname[256];
       //char fpath[56];
       //char program_version[60];
       char* dflt_iname=INFONAME_DFLT;   // information file default name
       time_t tstart, tfinish;    
//       mk5tok5_info_t mk5info;
       //gr_info_t gr; 
       k5m5runp_t runp;
       FILE *fo;
       //ip_vlbi_file_old_t k5[4];   // IP-VLBI data type variables 
       K5VSSP_file_t k5[4];   // IP-VLBI data type variables 
       m5head_t hd;
       int sync_pat=0xABADDEED;   // Mark5B sync pattern
       int imask[]={0,0x7FFFFFFF,0x3FFFFFFF,0,0xFFFFFFF,0,0,0,0xFFFFFF};
                 // masking table for 1,2,4,8 bit sampling data
       int *conv2bit;
       int conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)
       int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (default)
       int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       } 
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("        %s ---   default directory for Mark5 vex file\n",M5VEX);
          printf("                      ( %s )\n",getenv(M5VEX));
          printf("\n");
          exit(0);
          }
       }

       strcpy(runp.program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_k5tom5b(argc, argv, &runp); 

       if(runp.e2bit==1){
           conv2bit=conv2bit_1;
       } else if(runp.e2bit==3){
           conv2bit=conv2bit_3;
       } else {
           conv2bit=conv2bit_2;
       }

       if (runp.info_file[0]=='\0'){
          strcpy(runp.info_file,dflt_iname);
       }

       /* out directory create if not set */

       if(runp.mk5name[0] != 0 && runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.mk5name, runp.outdir, fname); // debug
          printf("A %s\n",runp.mk5name);
          printf("A %s\n",runp.outdir);
          printf("A %s\n",fname);
       }


       setup_k5files_full_path(&runp);  // added on 2008-09-30

       head=obuf;
       idata=obuf+4;

       adjust_flag=1;   // set period is integer second mode  2010-07-21

       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    K5/VSSP to Mark-5B Data Format Converter              *\n");
       printf("*      %s  by  T.KONDO/NICT    *\n",runp.program_version);
       if(runp.nocheck==1)
       printf("*               (!! No header check mode!!)                *\n");
       printf("************************************************************\n");
       printf("\n");

       //if(runp.monit==1)
//       monit_mk5info(&mk5info);  // debug    

       if(runp.monit==1) monit_runp(&runp);

       tstart= time(NULL);


       /* K5 data check to get sampling information */

       totch=0;
       bitsample=0;
       for(n=0; n<runp.k5num; n++){
          /* K5 type variable initialize 2007-04-24 */ 
          k5vssp_t_ini(&k5[n]);  // 2007-04-24
          //FileCloseFlag[n]=0;  // 2008-07-04
          K5_FileOpenFlag[n]=0;  // 2016-08-15
          /* Data file open */
          //k5[n].ibuf  = (unsigned char *)malloc(numb);             
          //k5[n].fname=runp.k5name[n];
          strcpy(k5[n].fname,runp.k5name[n]);
          if( (k5[n].stream  = fopen( runp.k5name[n], "rb" )) == NULL ){
              printf("Data File %s   Open Error! \n",runp.k5name[n]);
              goto exitend;
          }
          K5_FileOpenFlag[n]=1;  // 2016-08-15
          k5[n].size=get_file_size(k5[n].fname); // added on 2009-07-29
         /************************************************/
         /* Data header check to get sampling parameters */
         /************************************************/

         // printf("Data File %s \n",k5[n].fname);
         //if(chk_K5_header(&k5[n]) < 0){
          if(CheckK5header_with_Read(&k5[n]) < 0){  //2007-04-24
            goto exitend;
          }

          for(i=0; i<k5[n].numch; i++){  // set up K5 file ID and channel ID for K5 channel# (1-16) 
              K5file_id[totch+i]=n;       // 0-3
              K5chan_id[totch+i]=i;       // 0-3
          }

          totch+=k5[n].numch;
          bitsample+=k5[n].numch*k5[n].adbit;
          //printf("%d VSSP format %d\n",k5[n].seconds,k5[n].format);  // debug

          /* add 2010-07-21 */
          dwork=1000.0*(double)k5[n].sfreq*(double)(k5[n].adbit*k5[n].numch)/8.0; // bytes/sec without header
          dwork+=(double)k5[n].headsizew*4.0;  // bytes/sec including header
          sec_period[n]=(double)k5[n].size/dwork;   // data length in second frame unit
          isec_period[n]=(int)sec_period[n];

          printf("Second Length :  %f\n",sec_period[n]); // debug
          /* add end */


       }


       /* check date info is given or not for VSSP32 or VSSP64 format */  // added on 2016-08-15
       if(runp.mjd3==0 && runp.yyyy==0){
           for(n=0; n<runp.k5num; n++){
               if(k5[n].format > 1){  //  VSSP32 or VSSP64
                   runp.yyyy=k5[n].year;
                   runp.ddd =k5[n].totalday;
                   //get month day
                   ddd2md(runp.yyyy,runp.ddd,&runp.mm,&runp.dd);
                   //get MJD
                   runp.mjd=ut2mjd(&runp.dw,runp.yyyy,runp.mm,runp.dd,0,0,0);
                   runp.mjd3 = runp.mjd % 1000;
                   break;
               }
           }
       }
       if(runp.yyyy==0){
           printf("k5tom5b: Warning!!  Date information is not given!!\n");
           printf("       : Please run again with date information\n");
           goto exitend;
       }


        /********************************/
        /* set internal-use parameters  */
        /********************************/

        sfreq=k5[0].sfreq;       // sampling frequency (kHz)
        adbit=k5[0].adbit;       // A/D bits
        numch=k5[0].numch;       // # of channel in data (1 or 4)
        fsampl=1000.0*(double)sfreq;   // sampling frequency (Hz)
        tsampl=1.0/fsampl;            // sampling period (sec)
        bits=((int)fsampl)*adbit*numch;  // # of bits in 1 sec data except header
        bytes=bits/8;                     // # of bytes in 1 sec data except header

        samplpp=(int)fsampl;  // # of samples per ch in a second


        if(adbit > 2){
           printf("AD bits (%d) exceeds 2 bits!\n",adbit); 
           printf("So data are converted to Mark 5B format!\n"); 
           goto exitend;
        }

        if( (32 % bitsample) !=0) {
           printf(" # of bits per samples is %d\n",bitsample); 
           printf(" So that # of samples in a 32-bit word is not integer!\n"); 
           printf(" This situation is not allowded in the current version.\n");
           goto exitend;
        }
        nsample32=32/bitsample;  // number of samples per ch in a 32-bit word 

        //debug
        //printf("samplpp,totch, bitsample,nsample32  %d %d %d %d\n",
        //  samplpp, totch,bitsample, nsample32);




        /*****************************************/
        /*  set unit # of samples to process     */
        /*****************************************/

        //hd.fanout=1;  //  fan out in Mark 5 
        //fanout=4;  //  fan out in Mark 5 (for Huygens)
        //usampl=2500*nsample32;   // # of samples in a Mark5B frame
        usampl=frame32*nsample32;   // # of samples per ch in a Mark5B frame
                        // Mark5B data frame ==> 2500 X 32 bits
                        //        header     ==>    4 X 32 bits



        //imax=(sfreq*1000)/(usampl*fanout);        // # of usampls in a second
        imax=(sfreq*1000)/usampl;        // # of Mrk5B frames in a second

        numb=usampl*adbit*numch;  // # of bits in a usampl per K5 file
        //printf("# of bits per usampl  %d\n",numb);
        numb/=32;  // # of 32-bit word in a usampl per K5 file

        unitp=(double)usampl*tsampl;   // period in a usampl (sec)

        for(n=0; n<runp.k5num; n++){
           k5[n].ubuf  = (unsigned int *)malloc(numb*sizeof(int));
           k5[n].idata = (int *)malloc(usampl*numch*sizeof(int));
           k5[n].bytesec=bytes;     // 2009-07-29
        }

        //printf("usampl,imax, numb, unitp  %d %d %d %f\n",usampl,imax,numb,unitp);




       // if start time was different each file, then atamadashi is necessary

       /* check start time */
       aflag=0;   // atamadashi flag
       secmin=k5[0].seconds;
       secmax=k5[0].seconds;
       if(runp.k5num > 0) {
          for(n=1; n<runp.k5num; n++){
              if(k5[n].seconds < secmin) secmin=k5[n].seconds;
              if(k5[n].seconds > secmax) secmax=k5[n].seconds;
          }
          if(secmin != secmax){
              aflag=1;
              printf(" Warning!! start time different file exists!\n");
          }
       }

       if(aflag==1){ // atamadashi at secmax
           for(n=0; n<runp.k5num; n++){
              isa= secmax-k5[n].seconds;
              isec_period[n]-=isa;  // 2010-07-21
              printf("atamadashi %d\n",isa);
              //if(fwd_K5_Nsec(k5[n].ibuf, isa, numb, imax, &k5[n]) <0){
              if(fwdNsecK5(isa, &k5[n]) <0){  //2007-04-24
                goto exitend;
              }
           }
       }

       // find common data length in sec 2010-07-21
       common_sec_length=isec_period[0];
       for(n=0; n<runp.k5num; n++){
           //if(common_sec_length < isec_period[n]) common_sec_length=isec_period[n];
           if(common_sec_length > isec_period[n]) common_sec_length=isec_period[n];  // 2012-12-03
       }


       // start offset
       if(runp.soffset > 0){
           for(n=0; n<runp.k5num; n++){
              isa= runp.soffset;
              printf("File %d forward %d sec    ",n+1,isa);
              //if(fwd_K5_Nsec(k5[n].ibuf, isa, numb, imax, &k5[n]) <0){
              if(fwdNsecK5(isa, &k5[n]) <0){  // 2007-04-24
                goto exitend;
              }
           }
           common_sec_length-=runp.soffset; // 2010-07-21
           printf("\n");
       }


       // set actual data conversion period if set so  (2010-07-21)
       if(adjust_flag==1){
          if(runp.period > 0 ) {
             if( runp.period > common_sec_length) runp.period=common_sec_length;
          } else {
             runp.period=common_sec_length;
          }
          if(runp.period == 0){
              printf(" K5 data length is less than 1 sec. So conversion to Mark-5B is cancelled!\n");
              goto exitend;
          }
       } 


       /* Mark5 file create */

       M5_FileOpenFlag=0;    // 2016-08-15

       if(runp.mk5name[0]=='\0'){
           printf(" Mark 5 file not created!\n");
           goto exitend;

       } else {
          if( (fo  = fopen( runp.mk5name, "wb" )) == NULL ){
              printf("  File  %s  Open Error! \n",runp.mk5name);
              goto exitend;
          }
          M5_FileOpenFlag=1;  // 2016-08-15
       }

       /* Mark5 header common info setup */

       //hd.yyyy=runp.yyyy;
       //hd.ddd=runp.ddd;
       //hd.mjd=runp.mjd3;
       mjd=runp.mjd3;
       dsec=0.0; //2011-01-09
       dmsec10=0.0;
       msec10=0;
       seconds_init=k5[0].seconds;  // 2011-01-09
       seconds=seconds_init;        // 2011-01-09
       sec_count=0;   // 2012-12-03
       tvgflag=0;
       uid=0xF00F;

       //hd.nbits=mk5info.ntrack;  // bits per word.   now supports 16, 32 and 64 bit modes

       //hd.numch=mk5info.nch;  // # of Mark5 channels


       /* conversion start message */

       printf("\n");
       printf("K5 file(s)\n");
       k=1;  // K5 ch# start
       for(n=0; n<runp.k5num; n++){
           if(k5[n].numch==1){ // 1ch mode data
               printf(" FILE%d CH# %2d    : %s\n",n+1,k, k5[n].fname);
               k++;
           } else {  //4ch mode
               printf(" FILE%d CH# %2d-%2d : %s\n",n+1,k,k+3, k5[n].fname);
               k+=4;
           }
           printf("   (A/D(bits) %d  CHs %d  SFreq(kHz) %d",
                     k5[n].adbit,k5[n].numch,k5[n].sfreq);
           printf(" Time %02d:%02d:%02d  sec %d)",
                      k5[n].hh,k5[n].mm,k5[n].ss,k5[n].seconds);
           if(k5[n].format==1){                 // add 2007-04-24
               printf(" VSSP\n");
           } else if(k5[n].format==2){
               printf(" VSSP32\n");
           } else {
               printf("\n");
           }
       }

       printf("\n");

       printf("Mark5B file (created) : %s\n",runp.mk5name);
       printf("    Observation Date : 3digit MJD = %4d\n",runp.mjd3);
       if(runp.yyyy != 0){
       printf("                     :      Year  = %4d\n",runp.yyyy);
       printf("                     :      Month = %4d\n",runp.mm);
       printf("                     :      Day   = %4d\n",runp.dd);
       printf("                     :  Total Day = %4d\n",runp.ddd);
       }
       if(adjust_flag ==1){
       printf("    Conversion length (sec) :  %d\n",runp.period);
       }
       printf("\n");

       hd.vlba=1;
       printf("                   TIME LABEL MONITOR\n");
       if(hd.vlba==1){
       printf(" ==========================================================\n");   
       printf("  K5-FILE1  K5-FILE2  K5-FILE3  K5-FILE4    VLBA-TIME\n");   
       printf("  HH:MM:SS  HH:MM:SS  HH:MM:SS  HH:MM:SS  JJJSSSSSssss\n");   
       printf(" ----------------------------------------------------------\n");   
       } else {
       printf(" ==========================================================\n");   
       printf("  K5-FILE1  K5-FILE2  K5-FILE3  K5-FILE4   MarkIV-TIME\n");   
       printf("  HH:MM:SS  HH:MM:SS  HH:MM:SS  HH:MM:SS  YDDDHHMMSSsss\n");   
       printf(" ----------------------------------------------------------\n");   
       }

//       tmax=k5[0].seconds;
//       tmin=tmax;
//       for(n=0; n<runp.k5num; n++){
//           printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
//           if(k5[n].seconds > tmax) tmax=k5[n].seconds;
//           if(k5[n].seconds < tmin) tmin=k5[n].seconds;
//       }
//       if(runp.k5num != 4){
//           for(n=runp.k5num; n<4; n++) printf("          ");
//       }
//       if(hd.vlba==1){
//           printf("  %03d%05d%04d",mjd,seconds,msec10);
//       } else {
//           printf("  %1d%03d%02d%02d%02d%03d",
//               (runp.yyyy % 10),runp.ddd,(int)(seconds/3600),
//               (int)((seconds%3600)/60),(seconds%60),
//               msec10/10);
//       }
//       if(tmax==tmin){
//           printf("  OK\n");
//       } else {
//           printf("  ERROR (TIME DIFFER!!)\n");
//           exit(1);
//       }

       period=0;


       ow=4;

       /* mark5B header generation */
       //m5b_hd_make(head, mjd, seconds, msec10,
       //        uid, tvgflag, nframe);

       //printf("Debug imax, koffset %d %d\n",imax,koffset);

       while(1){    // entry point of second loop

           nframe=0;  // frame # (in a second) counter reset


           for(i=0; i<imax; i++){  // 1 sec loop is imax'th Mark5 frame

               /* 2012-12-03 new algorythm applied and moved from below */
               // time calc for Mark5 header
               dsec=(double)nframe*unitp;   // fractional second
               dmsec10=dsec*10000.0;
               msec10=(int)dmsec10;
               seconds=seconds_init+sec_count;
               mjd=runp.mjd3+(int)((double)seconds/86400.0);
               if(seconds >= 86400) seconds-=86400;          // 2016-07-20

               if(i==0){ //  print time label  2012-12-03
                  tmax=k5[0].seconds;
                  tmin=tmax;
                  for(n=0; n<runp.k5num; n++){
                      printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
                      if(k5[n].seconds > tmax) tmax=k5[n].seconds;
                      if(k5[n].seconds < tmin) tmin=k5[n].seconds;
                  }
                  if(runp.k5num != 4){
                      for(n=runp.k5num; n<4; n++) printf("          ");
                  }

                  printf("  %03d%05d%04d",mjd,seconds,msec10);
                  if(msec10 != 0){
                      printf("  ERROR (TIME DIFFER!!)\n");
                  } else {
                      if(tmax==tmin){
                          if(seconds==k5[0].seconds){
                             printf("  OK\n");
                          } else {
                             if(runp.nocheck==1){ // 2010-08-03
                                printf("  TIME ERROR IGNORED!!\n");
                             } else {
                                printf("  ERROR (TIME DIFFER!!)\n");
                             }
                          }
                       } else {
                          if(runp.nocheck==1){ // 2010-08-03
                             printf("  TIME ERROR IGNORED!!\n");
                          } else {
                             printf("  ERROR (TIME DIFFER!!)\n");
                          }
                       }
                   }
			   }

               /* read K5 data corresponding to one frame of Mark 5B */ 
               for(n=0; n<runp.k5num; n++){
                  //iread = fread( k5[n].ibuf, sizeof( char ), numb, k5[n].stream);
                  iread = fread( k5[n].ubuf, sizeof(int), numb, k5[n].stream);
                  if (iread <= 0) goto exit_loop;  // exit loop  2008-07-04
                       // all K5 file must be the same sampling condition
                  parseADdataKOKO(k5[n].ubuf,numb,adbit,k5[n].idata, &kpos);
                  //parseADdata_3N(&k5[n]);
               }
               //kpos=k5[0].kpos;

               //printf("Debug numb kpos %d %d\n",numb,kpos);

               /* convert to Mark5 frame data */

               // mark5B header generation 

               /* 2011-01-09 */  // obsoleted on 2012-12-03 because of enough accuracy at dmsec10
               //dsec+=unitp;
               ////dmsec10=dsec*10000.0+0.5;   // "+0.5" is for round-off
               //dmsec10=dsec*10000.0;   // "+0.5" is for round-off
               //msec10=(int)fmod(dmsec10,10000.0);
               //seconds=seconds_init+(int)(dmsec10/10000.0);
               //mjd=runp.mjd3+(int)(dmsec10/(86400.0*10000.0));


               
               //printf("msec10  %d\n",msec10);
               m5b_hd_make(head, mjd, seconds, msec10,
                           uid, tvgflag, nframe);

               // frame data generation
               ow=4;
#ifdef ORIGINAL
               for(k=0; k<frame32; k++){
                   for(m=0; m<nsample32; m++){
                      for(n=0; n<runp.k5num; n++){
                          for(ch=0; ch< k5[n].numch; ch++){
                              iw=k5[n].idata[ch+k5[n].numch*(k*nsample32+m)];
                              if(adbit==2) iw=conv2bit[iw]; // code conversion in case of 2 bit
                              iw <<= (32-adbit);
                              idata[k]=((idata[k] >> adbit) & imask[adbit]) | iw;
                          }
                      }
                   }
                   ow++;
               }
#else
               // new procedure     2014-05-24
               for(k=0; k<frame32; k++){
                   for(m=0; m<nsample32; m++){
                       for(ch=0; ch<totch; ch++){
                           f_id=K5file_id[runp.ch_assign[ch]-1];
                           ch_id=K5chan_id[runp.ch_assign[ch]-1];
                           iw=k5[f_id].idata[ch_id+k5[f_id].numch*(k*nsample32+m)];
                           if(adbit==2) iw=conv2bit[iw]; // code conversion in case of 2 bit
                           iw <<= (32-adbit);
                           idata[k]=((idata[k] >> adbit) & imask[adbit]) | iw;
                       }
                   }
                   ow++;
               }
#endif
               //printf("debug1 nbuf k adbit %d %d %d\n",nbuf,k,adbit);
               iwrite=fwrite(obuf,sizeof(int),nbuf,fo);
               //printf("debug2 \n");

               /* time info update */

               //dmsec10+=(unitp*10000.0);
               //msec10=(int)(dmsec10+0.5);
               //if(msec10 >= 10000){
               //   msec10-=10000; 
               //   dmsec10=(double)msec10;
               //   seconds++;
               //   if(seconds >= 86400){
               //       seconds-=86400;
               //       mjd++;
               //   }
               //}

               /* 2011-01-09 */  // obsoleted on 2012-12-03 because of enough accuracy at dmsec10
               //dsec+=unitp;
               //dmsec10=dsec*10000.0+0.5;   // "+0.5" is for round-off
               //msec10=(int)fmod(dmsec10,10000.0);
               //seconds=seconds_init+(int)(dmsec10/10000.0);
               //mjd=runp.mjd3+(int)(dmsec10/(86400.0*10000.0));

               /* 2012-12-03 new algorythm applied and this should be in front of Mark5 header generation */
               //dsec=(double)nframe*unitp;   // fractional second
               //dmsec10=dsec*10000.0;
               //msec10=(int)dmsec10;
               //seconds=seconds_init+sec_count;
               //mjd=runp.mjd3+(int)((double)seconds/86400.0);


               /* frame # update */

               nframe++;

           }
           sec_count++;  //2012-12-03

           // second data read end

           period++;
           if(runp.period > 0 && (runp.period == period)) goto exit_loop;

           // K5 header read
           tmax=k5[0].seconds;
           tmin=tmax;
           for(n=0; n<runp.k5num; n++){
              if(runp.nocheck==1){  // 2010-08-03
                 if(CheckK5header_with_Read_but_no_search(&k5[n])<0){ //2010-08-03
                     FileCloseFlag[n]=1;
                     goto exit_loop;  // already file is closed
                 }
              } else {
                 if(CheckK5header_with_Read(&k5[n]) < 0){  //2007-04-24
                     FileCloseFlag[n]=1;
                     goto exit_loop;  // already file is closed
                 }
              }
//              if(n==0){
//                 tmax=k5[0].seconds;
//                 tmin=tmax;
//              }
//              printf("  %02d:%02d:%02d",k5[n].hh,k5[n].mm,k5[n].ss);
//             if(k5[n].seconds > tmax) tmax=k5[n].seconds;
//              if(k5[n].seconds < tmin) tmin=k5[n].seconds;
           } 
//           if(runp.k5num != 4){
//              for(n=runp.k5num; n<4; n++) printf("          ");
//           }
//
//           printf("  %03d%05d%04d",mjd,seconds,msec10);
//           if(msec10 != 0){
//              printf("  ERROR (TIME DIFFER!!)\n");
//           } else {
//              if(tmax==tmin){
//                  if(seconds==k5[0].seconds){
//                     printf("  OK\n");
//                  } else {
//                      if(runp.nocheck==1){ // 2010-08-03
//                          printf("  TIME ERROR IGNORED!!\n");
//                      } else {
//                          printf("  ERROR (TIME DIFFER!!)\n");
//                      }
//                  }
//              } else {
//                 if(runp.nocheck==1){ // 2010-08-03
//                     printf("  TIME ERROR IGNORED!!\n");
//                 } else {
//                     printf("  ERROR (TIME DIFFER!!)\n");
//                 }
//              }
//           }
       }

exit_loop:
exitend:

       printf(" ==========================================================\n");   

       for(n=0; n<runp.k5num; n++){
           //if(FileCloseFlag[n]==0) i=fclose(k5[n].stream);
           if(K5_FileOpenFlag[n]==1) i=fclose(k5[n].stream);
       }

       if(M5_FileOpenFlag==1){    // 2016-08-15

           if(ow < nbuf){
                iwrite=fwrite(obuf,sizeof(int),ow,fo);
           }

           i=fclose(fo);


           printf("\n");
           printf("  Mark5 file (%s) has been created.\n\n",runp.mk5name);
           tfinish= time(NULL);
           telapsed = difftime(tfinish,tstart);
           printf("Time elapsed (sec) for One obs process is %f\n",telapsed);
       }

//exitend: 
       return 0;
}
